import java.security.InvalidAlgorithmParameterException;
import java.security.InvalidKeyException;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.security.SecureRandom;
import java.security.spec.AlgorithmParameterSpec;
import java.security.spec.InvalidKeySpecException;
import java.security.spec.KeySpec;
import java.util.Base64;
import java.util.Base64.Encoder;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.math.BigInteger;
import java.nio.file.Files;
import java.nio.file.Paths;

import javax.crypto.BadPaddingException;
import javax.crypto.Cipher;
import javax.crypto.IllegalBlockSizeException;
import javax.crypto.NoSuchPaddingException;
import javax.crypto.SecretKey;
import javax.crypto.SecretKeyFactory;
import javax.crypto.spec.GCMParameterSpec;
import javax.crypto.spec.PBEKeySpec;
import javax.crypto.spec.SecretKeySpec;

public class Cryptor
{

	private char[] alphabet =
		{ 'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V',
				'W', 'X', 'Y', 'Z', 'Ä', 'Ö', 'Ü' };

	// ******************* Aufgabe 1 Beginn
	// *****************************************
	/**
	 * En- or decrypts the given inString with the given key based on mode
	 * 
	 * @param inString: The string to be en- or decrypted
	 * @param key:      The key use for en- or decryption
	 * @param mode:     Either 1 or -1. 1 selects encryption and -1 selects
	 *                  decryption
	 * @return The en- or dectypted text
	 */
	public String cryptCesar(String inString, int key, int mode)
	{
		StringBuilder translatedBuild = new StringBuilder();

		if (mode != -1 && mode != 1)
			mode = 1;

		for (char c : inString.toUpperCase().toCharArray())
		{
			int  x = new String(alphabet).indexOf(c);
			char z = c;
			if (x != -1)
				z = alphabet[(x + mode * key + alphabet.length) % alphabet.length];
			translatedBuild.append(z);
		}

		return translatedBuild.toString();
	}

	/**
	 * En- or decrypts the given inString with the given key based on mode
	 * 
	 * @param inString: The string to be en- or decrypted
	 * @param key:      The key use for en- or decryption
	 * @param mode:     Either 1 or -1. 1 selects eincryption and -1 selects
	 *                  decryption
	 * @return The en- or dectypted text
	 */
	public String cryptVigenere(String inString, String key, int mode)
	{
		StringBuilder translatedBuild = new StringBuilder();

		if (mode != -1 && mode != 1)
			mode = 1;
		if (key.isBlank())
			return "Kein key angegeben?";

		for (int i = 0; i < inString.toUpperCase().toCharArray().length; i++)
		{
			char c = inString.toUpperCase().toCharArray()[i];

			int  x = new String(alphabet).indexOf(c);
			int  y = new String(alphabet).indexOf(key.toUpperCase().toCharArray()[(i + key.length()) % key.length()]);
			if (y == -1)
			{
				translatedBuild.append(c);
				continue;
			}
//				return "Der key darf keine Sonderzeichen enthalten!";

			char z = c;
			if (x != -1)
				z = alphabet[(x + mode * y + alphabet.length) % alphabet.length];

			translatedBuild.append(z);
		}

		return translatedBuild.toString();

	}
	// ******************* Aufgabe 1 Ende *****************************************

	// ******************* Aufgabe 2 Beginn
	// *****************************************
	/**
	 * Brute-force the encrypted string.
	 * 
	 * @return All possible decrypted texts delimitted by a newline.
	 */
	public String bruteForceCesar()
	{
		String        encryptedMessage = "SXT ITÄF XEF PÄÄTE, IPE STD UPÄÄ XEF. SXT ITÄF XEF SXT VTEPÖFWTXF STD FPFEPRWTÜ, ÜXRWF STD SXÜVT. SXT ITÄF XEF SGDRW SXT FPFEPRWTÜ QTEFXÖÖF GÜS SPSGDRW, SPEE TE PÄÄT FPFEPRWTÜ EXÜS. STÜÜ, SXT VTEPÖFWTXF STD FPFEPRWTÜ QTEFXÖÖF, IPE STD UPÄÄ XEF GÜS PGRW, IPE PÄÄTE ÜXRWF STD UPÄÄ XEF. SXT FPFEPRWTÜ XÖ ÄAVXERWTÜ DPGÖ EXÜS SXT ITÄF. SXT ITÄF LTDUMÄÄF XÜ FPFEPRWTÜ. TXÜTE ZPÜÜ STD UPÄÄ ETXÜ ASTD ÜXRWF STD UPÄÄ ETXÜ GÜS PÄÄTE OQDXVT VÄTXRW QÄTXQTÜ. IPE STD UPÄÄ XEF, SXT FPFEPRWT, XEF SPE QTEFTWTÜ HAÜ EPRWHTDWPÄFTÜ. STD EPRWHTDWPÄF XEF TXÜT HTDQXÜSGÜV HAÜ VTVTÜEFMÜSTÜ. (EPRWTÜ, SXÜVTÜ.) TE XEF STÖ SXÜV ITETÜFÄXRW, STD QTEFPÜSFTXÄ TXÜTE EPRWHTDWPÄFTE ETXÜ LG ZNÜÜTÜ. XÜ STD ÄAVXZ XEF ÜXRWFE LGUMÄÄXV: ITÜÜ SPE SXÜV XÖ EPRWHTDWPÄF HADZAÖÖTÜ ZPÜÜ, EA ÖGEE SXT ÖNVÄXRWZTXF STE EPRWHTDWPÄFTE XÖ SXÜV QTDTXFE BDMYGSXLXTDF ETXÜ.";
		StringBuilder translatedBuild  = new StringBuilder();

		for (int i = 0; i < alphabet.length; i++)
		{
			translatedBuild.append(i + ": " + cryptCesar(encryptedMessage, i, -1) + "\n");
		}

		return translatedBuild.toString();
	}

	/**
	 * Decrypt cesar encrypted string using the letter-frequency method
	 * 
	 * @return The key used for encryption
	 */
	public int letterfreqCesar()
	{
		String                  encryptedMessage = "WÖX MXBJ ÖIJ TBBXI, MTI WXH YTBB ÖIJ. WÖX MXBJ ÖIJ WÖX ZXITCJÄXÖJ WXH JTJITVÄXD, DÖVÄJ WXH WÖDZX. WÖX MXBJ ÖIJ WKHVÄ WÖX JTJITVÄXD UXIJÖCCJ KDW WTWKHVÄ, WTII XI TBBX JTJITVÄXD IÖDW. WXDD, WÖX ZXITCJÄXÖJ WXH JTJITVÄXD UXIJÖCCJ, MTI WXH YTBB ÖIJ KDW TKVÄ, MTI TBBXI DÖVÄJ WXH YTBB ÖIJ. WÖX JTJITVÄXD ÖC BEZÖIVÄXD HTKC IÖDW WÖX MXBJ. WÖX MXBJ PXHYQBBJ ÖD JTJITVÄXD. XÖDXI ATDD WXH YTBB IXÖD EWXH DÖVÄJ WXH YTBB IXÖD KDW TBBXI SUHÖZX ZBXÖVÄ UBXÖUXD. MTI WXH YTBB ÖIJ, WÖX JTJITVÄX, ÖIJ WTI UXIJXÄXD LED ITVÄLXHÄTBJXD. WXH ITVÄLXHÄTBJ ÖIJ XÖDX LXHUÖDWKDZ LED ZXZXDIJQDWXD. (ITVÄXD, WÖDZXD.) XI ÖIJ WXC WÖDZ MXIXDJBÖVÄ, WXH UXIJTDWJXÖB XÖDXI ITVÄLXHÄTBJXI IXÖD PK ARDDXD. ÖD WXH BEZÖA ÖIJ DÖVÄJI PKYQBBÖZ: MXDD WTI WÖDZ ÖC ITVÄLXHÄTBJ LEHAECCXD ATDD, IE CKII WÖX CRZBÖVÄAXÖJ WXI ITVÄLXHÄTBJXI ÖC WÖDZ UXHXÖJI FHQÜKWÖPÖXHJ IXÖD.";
		int                     key              = 0;

		/**
		 * Zählt, wie oft die chars in encryptedMessage vorkommen Source:
		 * https://stackoverflow.com/a/6712708
		 */
		Map<Character, Integer> map              = new HashMap<Character, Integer>();
		for (int i = 0; i < encryptedMessage.length(); i++)
		{
			char    c   = encryptedMessage.charAt(i);
			Integer val = map.get(c);
			if (val != null && new String(alphabet).indexOf(c) != -1)
			{
				map.put(c, val + 1);
			}
			else if (new String(alphabet).indexOf(c) != -1)
			{
				map.put(c, 1);
			}
		}

		/**
		 * Welcher char am meisten vorkommt Source: https://stackoverflow.com/a/28015042
		 */
		char x = Collections.max(map.entrySet(), Map.Entry.comparingByValue()).getKey();

		key = new String(alphabet).indexOf(x) - new String(alphabet).indexOf('E');

		return key;
	}

	/**
	 * Use "known plaintext" to decrypt vigenere encrypted text
	 * 
	 * @return The key used for encryption
	 */
	public String knownPlaintextVigenere()
	{
		String encrypted = "IVHI ZVEÖBJV XNCVN EDU ÖRHFEE, GIY ÄRBYÄ VS ÄKB ÄRITHUSYH XRD 'LÄUFEWUURRYÄ TOXR' WÜR UEI LZGYÄXFE PEUE NKTH TXBAEÄJ WÜBW MÜDJVLJ 'ÖNFGD PCNÖBTYHJ' UÄZFIZS HAKCQTHCVVY QE SRYPÄVN. ZFAKEBZEIR. EÜÄXF UÄIHECYKBG OXZ XRH NJN CDYB RBYB UED RGTYBHVIWUITHYÄ GHHTWEW CJVHK DIN ÄÖTHK CVHI ÖA NRZV. GÖV JVX SYUXB ÜCJ NÜPÄHS EDSRYPÄSAI. GEÜRD SÜR RCÖDIRM GXÜCÖR UAKRD SÜR KIY ZVSÜPÄVRK MCHÜÄ QBYBJFAÄRD. XRDB 'ÄKF NRÖÜ ÜPÄ PUBTBOÜQ SIE. YEÜCIH XNI NÜPÄH XNIG ÜPÄ NÜPÄH MRHWOCTJ WYBWV' (PI ÖLHVN NKTH VC SFÄZ VFÄ BIIFTBA - HEIBÖHOIVTÜ GVIGIETI - UJHPJ://WNG.MOLDKSE.PEA/NNJTH?F=SM6VP-I5ÖCE - TS 1:26). ÖB XVXGED IZNER KÜECVYE FÖVL CFRSJ CZT QXF BBODTFTHRPÖVX UEQ UED UFEWUXB MÖD CFQXG. ZÖH ZBXINXYÖTHYÄ XRTCIVN NDCNPZEI";
		String opener    = "Sehr geehrte Damen und Herren";
		String end       = "Mit freundlichen Grüssen Anonymous";
		String key       = "";
		// ZÖH ZBXINXYÖTHYÄ XRTCIVN NDCNPZEI
		// ÄZOZ CZIFSÜZ ÄNOZÄ YQX FVYWIR
//		key = cryptVigenere(encrypted, opener, 1);

//		key = cryptVigenere(opener, encrypted, 1);

		/*
		 * Nicht gelöst
		 */

		return key;
	}
	// ******************* Aufgabe 2 Ende *****************************************



	// ******************* Aufgabe 3 Beginn
	// *****************************************
	public String hash(String hasher, String input)
	{
		if (input.isBlank())
			return "Kein Input angegeben.";
		String        hashString = "";
		MessageDigest digest     = null;
		switch (hasher)
			{
			case "MD5":
				try
				{
					digest = MessageDigest.getInstance("MD5");
				} catch (NoSuchAlgorithmException e)
				{
					System.err.println("No such Algorithm");
				}
				break;
			case "SHA":
				try
				{
					digest = MessageDigest.getInstance("SHA-1");
				} catch (NoSuchAlgorithmException e)
				{
					System.err.println("No such Algorithm");
				}
				break;
			case "SHA256":
				try
				{
					digest = MessageDigest.getInstance("SHA-256");
				} catch (NoSuchAlgorithmException e)
				{
					System.err.println("No such Algorithm");
				}
				break;
			case "SHA512":
				try
				{
					digest = MessageDigest.getInstance("SHA-512");
				} catch (NoSuchAlgorithmException e)
				{
					System.err.println("No such Algorithm");
				}
				break;
			case "SHA3":
				try
				{
					digest = MessageDigest.getInstance("SHA3-512");
				} catch (NoSuchAlgorithmException e)
				{
					System.err.println("No such Algorithm");
				}
				break;
			default:
				return "Keine gültige Hashfunktion angegeben.";
			}

		byte mdByteData[] = digest.digest(input.getBytes());

		hashString = byteArrayToHexString(mdByteData);
		return hashString;
	}

	public boolean areEqual(String filePath1, String filePath2)
	{
		try
		{
			byte[]        firstFile  = Files.readAllBytes(Paths.get(filePath1));
			byte[]        secondFile = Files.readAllBytes(Paths.get(filePath2));

			MessageDigest digest     = MessageDigest.getInstance("SHA-512");
			String        hashFile1  = byteArrayToHexString(digest.digest(firstFile));
			String        hashFile2  = byteArrayToHexString(digest.digest(secondFile));

			if (hashFile1.equals(hashFile2))
				return true;

			return false;
		} catch (Exception e)
		{
			System.out.println(e.toString());
			return false;
		}

	}
	// ******************* Aufgabe 3 Ende *****************************************

	// ******************* Aufgabe 4 Anfang
	// *****************************************
	public String encryptAEAG(char[] password, String cleartext)
	{
		// TODO: implement AES authenticaed encryption
		int           i           = 16000;
		StringBuilder output      = new StringBuilder();

		SecureRandom  random      = new SecureRandom();
		byte          saltBytes[] = new byte[16];       // 128 bits are converted to 16 bytes;
		random.nextBytes(saltBytes);
		Encoder encoder = Base64.getUrlEncoder().withoutPadding();
		String  salt    = encoder.encodeToString(saltBytes);
		output.append(salt + "\n");

		output.append(i + "\n");


		try
		{
			SecretKey secretKey = encryptSecretKey(password, saltBytes, i);

			byte      iv[]      = new byte[12];
			random.nextBytes(iv);
			output.append(encoder.encodeToString(iv) + "\n");

			Cipher           cipher        = Cipher.getInstance("AES/GCM/NoPadding");
			GCMParameterSpec parameterSpec = new GCMParameterSpec(128, iv);          // 128 bit auth tag length
			cipher.init(Cipher.ENCRYPT_MODE, secretKey, parameterSpec);

			byte[] cipherText = cipher.doFinal(cleartext.getBytes());
			output.append(encoder.encodeToString(cipherText) + "\n");

//			System.out.println(output);

		} catch (NoSuchAlgorithmException | NoSuchPaddingException | InvalidKeyException
				| InvalidAlgorithmParameterException | IllegalBlockSizeException | BadPaddingException
				| InvalidKeySpecException e)
		{
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

		return output.toString();
	}

	private SecretKey encryptSecretKey(char[] password, byte[] saltBytes, int i)
			throws NoSuchAlgorithmException, InvalidKeySpecException
	{
		SecretKeyFactory keyFactory = SecretKeyFactory.getInstance("PBKDF2WithHmacSHA256");
		KeySpec          keySpec    = new PBEKeySpec(password, saltBytes, i, 128);
		
		return new SecretKeySpec(keyFactory.generateSecret(keySpec).getEncoded(), "AES");
	}

	public String decryptAEAG(char[] password, String ciphertext)
	{
		// TODO: implement AES authenticaed decryption
		String split[] = ciphertext.split("\n");
		byte[] plainText = null;

		try
		{
			Cipher                 cipher = Cipher.getInstance("AES/GCM/NoPadding");

			AlgorithmParameterSpec gcmIv  = new GCMParameterSpec(128, Base64.getUrlDecoder().decode(split[2]));

			cipher.init(Cipher.DECRYPT_MODE, encryptSecretKey(password, Base64.getUrlDecoder().decode(split[0]), Integer.parseInt(split[1])),
					gcmIv);
			
			plainText = cipher.doFinal(Base64.getUrlDecoder().decode(split[3]));

		} catch (NoSuchAlgorithmException | NoSuchPaddingException | InvalidKeyException | NumberFormatException
				| InvalidAlgorithmParameterException | InvalidKeySpecException | IllegalBlockSizeException
				| BadPaddingException e)
		{
			return "Entschluesselung fehlgeschlagen!";
		}

		return new String(plainText);
	}
	// ******************* Aufgabe 4 Ende *****************************************

	// ******************* Aufgabe 5 Anfang
	// *****************************************
	//
	// use a private helper function here e.g. hashPasswordSalted(mode, password,
	// salt)
	//
	private String hashPasswordSalted(char[] password, String salt)
	{
		StringBuilder output    = new StringBuilder();
		char[]        hashedPwd = password;

		for (int i = 1; i <= 100000; i++)
		{
			output = new StringBuilder();
			output.append("1:");
			output.append(salt + ":");
			output.append(i + ":");
			hashedPwd = hash("SHA512", salt + new String(hashedPwd)).toCharArray();
			output.append(hashedPwd);

//			System.out.println(new String(hashedPwd));
		}
//		System.out.println(output.toString());
		return output.toString();
	}

	public String hashPassword(int mode, char[] password)
	{
		StringBuilder output      = new StringBuilder();
		char[]        hashedPwd   = password;

		SecureRandom  random      = new SecureRandom();
		byte          saltBytes[] = new byte[16];       // 128 bits are converted to 16 bytes;
		random.nextBytes(saltBytes);
		Encoder encoder = Base64.getUrlEncoder().withoutPadding();
		String  salt    = encoder.encodeToString(saltBytes);

		if (mode == 2)
		{
			int              rounds = 5000;
			KeySpec          spec   = new PBEKeySpec(password, saltBytes, rounds, 512);
			SecretKeyFactory factory;
			try
			{
				factory = SecretKeyFactory.getInstance("PBKDF2WithHmacSHA1");

				byte[] hash = factory.generateSecret(spec).getEncoded();

				hashedPwd = byteArrayToHexString(hash).toCharArray();
			} catch (NoSuchAlgorithmException | InvalidKeySpecException e)
			{
				e.printStackTrace();
			}

			output = new StringBuilder();
			output.append("2:");
			output.append(salt + ":");
			output.append(rounds + ":");
			output.append(hashedPwd);

//				System.out.println(output.toString());
		}
		else
		{
			output.append(hashPasswordSalted(hashedPwd, salt));
		}

		return output.toString();
	}

	public boolean verifyPassword(char[] passwordToVerify, String hashedPassword)
	{
		String[] parts        = hashedPassword.split(":");
		String   salt         = parts[1];

		String   hashedResult = hashPasswordSalted(passwordToVerify, salt);
//		System.out.println(hashedResult);
//		System.out.println(hashedPassword);

		if (hashedResult.equals(hashedPassword))
			return true;
		else
			return false;
	}
	// ******************* Aufgabe 5 Ende *****************************************

	private static String byteArrayToHexString(byte[] hash)
	{
		StringBuffer hexString = new StringBuffer();

		for (int i = 0; i < hash.length; i++)
		{
			String hex = Integer.toHexString(0xff & hash[i]);
			if (hex.length() == 1)
				hexString.append('0');
			hexString.append(hex);
		}
		return hexString.toString();
	}
}

